﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;

using Microsoft.Xna.Framework.Graphics;

namespace XnaDevRu.Assets
{
	/// <summary>
	/// The AssetLibrary is the run-time component which loads managed
	/// objects from the library files
	/// </summary>
	/// <remarks>
	/// Derived classes can replace this default behavior to implement pack files or asset compression.
	/// </remarks>
	public class AssetLibrary : IAssetImporter, IAssetExporter
	{
		const int StreamBufferSize = 1024 * 4; // 4Kb

		private DirectoryInfo _rootDirectory;
		private LibraryWatcher _libWatcher;
		private Dictionary<string, IAssetImporter> _fileExtToImporterMap = new Dictionary<string, IAssetImporter>();
		private Dictionary<string, IAssetExporter> _fileExtToExporterMap = new Dictionary<string, IAssetExporter>();
		public bool UseFileCache = true;

		public AssetLibrary()
		{
			_rootDirectory = new DirectoryInfo(AppDomain.CurrentDomain.BaseDirectory);
			_libWatcher = new LibraryWatcher(this);

		}

		public AssetLibrary(string rootPath)
		{
			_rootDirectory = new DirectoryInfo(rootPath);
			_libWatcher = new LibraryWatcher(this);

			if (!_rootDirectory.Exists)
				throw new InvalidOperationException("Can`t find folder: " + rootPath);

		}

		public void RegisterImporter(IAssetImporter importer)
		{
			if (importer == null)
				return;

			importer.Library = this;

            string[] parts = importer.FileExtension.Split(';');

            for (int i = 0; i < parts.Length; i++)
            {
                if (string.IsNullOrEmpty(parts[i]))
                    continue;

                _fileExtToImporterMap[parts[i]] = importer;
            }
		}

		public void RegisterExporter(IAssetExporter exporter)
		{
			if (exporter == null)
				return;

			exporter.Library = this;

			_fileExtToExporterMap[exporter.FileExtension] = exporter;
		}

		/// <summary>
		/// Returns the full path of the given file
		/// </summary>
		/// <param name="fileName"></param>
		/// <returns></returns>
		public string GetFileFullPath(string fileName)
		{
			if (Path.IsPathRooted(fileName))
				return fileName;

			string result = Path.Combine(_rootDirectory.FullName, fileName);

			return result;
		}

		/// <summary>
		/// Returns the full path of the given file
		/// </summary>
		/// <param name="fileName"></param>
		/// <returns></returns>
		public string GetFileFullPath(string fileName, bool search)
		{
			if (!search)
				return GetFileFullPath(fileName);

			if (Path.IsPathRooted(fileName))
				return fileName;

			FileInfo[] files = _rootDirectory.GetFiles(fileName, SearchOption.AllDirectories);

			if (files.Length == 0)
				return fileName; // warning

			return files[0].FullName;
		}


		/// <summary>
		/// Loads asset from intermediate data format
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="file"></param>
		/// <returns></returns>
		public T Load<T>(string file) where T : Asset, new()
		{
            return Load<T>(GetAssetFileInfo(file), true);
		}

        public T Load<T>(AssetFile assetFile) where T : Asset, new()
        {
            return Load<T>(assetFile, true);
        }

        /// <summary>
        /// Loads asset from intermediate data format
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="file"></param>
        /// <returns></returns>
        public T Load<T>(string file, bool addToWatchList) where T : Asset, new()
        {
            AssetFile assetFile = GetAssetFileInfo(file);

            return Load<T>(assetFile, addToWatchList);


        }

        public T Load<T>(AssetFile assetFile, bool addToWatchList) where T : Asset, new()
        {
            T asset = null;

			// if extension != (*.asset) try to load from *.asset file first
			bool useCache = string.Compare(assetFile.FileExtension, this.FileExtension, true) != 0;

			// if extension == (*.asset) try to load from *.asset file first
			if (assetFile.FileExtension == ".asset")
				useCache = true;

            if (useCache && UseFileCache)
            {

                // if cached file exists and it is fresh enough then we will use it
                if (assetFile.CacheFreshEnough)
                    asset = this.LoadFromCache<T>(assetFile, addToWatchList);

                if (asset != null)
                    return asset;
            }

            // import asset using importer

            IAssetImporter importer = null;

            _fileExtToImporterMap.TryGetValue(assetFile.FileExtension, out importer);

            if (importer == null)
                return null; // we have no importer registered for given file type

            asset = importer.Load<T>(assetFile);

            // something was wrong during importing
            if (asset == null)
                return null;

            asset.SetDependency(this, assetFile.SourceAssetFileFullName);

            if (!string.IsNullOrEmpty(asset.Name))
            {
                // if this is "root" importer save result to the cache then
                if (!importer.HasParentImporter && UseFileCache)
                {
                    // temporary prevent watcher notifications for this asset file
                    _libWatcher.CurrentWorkingFileName = assetFile.CachedAssetFileFullName;

                    // save asset to cache (and add to the watch list)
                    this.SaveToCache<T>(asset, assetFile.CachedAssetFileFullName, false);

                    _libWatcher.CurrentWorkingFileName = null;
                }

            }

            WatchAsset(asset, addToWatchList);

            return asset;

        }

		private void WatchAsset(Asset asset, bool addToWatchList)
		{
			if (addToWatchList)
			{
				_libWatcher.AddAssetDependencyToWatch(asset);

				AssetDependencyWalker walker = new AssetDependencyWalker();

				asset.Accept(walker);

				for (int i = 0; i < walker.DependencyList.Count; i++)
				{
					_libWatcher.AddAssetDependencyToWatch(walker.DependencyList[i].Asset);
				}
			}
		}

		/// <summary>
		/// Saves asset to intermediate data format
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="asset"></param>
		/// <param name="file"></param>
		public void Save<T>(T asset, string file) where T : Asset, new()
		{
			Save<T>(asset, file, true);
		}

		/// <summary>
		/// Saves asset to intermediate data format
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="asset"></param>
		/// <param name="file"></param>
		public void Save<T>(T asset, string file, bool addToWatchList) where T : Asset, new()
		{
			if (asset == null)
				return;

			string ext = FileExtension; // if the given file name has no extension - use (*.asset) by default
			string assetName = Path.GetFileNameWithoutExtension(file);

			if (Path.HasExtension(file))
				ext = Path.GetExtension(file);

			string cachedAssetFile = file + FileExtension;
			string targetAssetFile = assetName + ext;

			// if extension == (*.asset) try to save to *.asset file
			if (string.Compare(ext, FileExtension, true) != 0)
			{
				if (UseFileCache)
					SaveToCache<T>(asset, cachedAssetFile, addToWatchList);
				
				return;
			}

			// export asset using exporter

			IAssetExporter exporter = null;

			_fileExtToExporterMap.TryGetValue(ext.ToLowerInvariant(), out exporter);

			if (exporter == null)
				return; // we have no exporter registered for given file type

			DateTime version = DateTime.Now;

			// if the version was not set by asset importer we will set it now
			if (asset.Version.Ticks == 0L)
				asset.Version = version;

			exporter.Save<T>(asset, targetAssetFile);

			SetStreamVersion(targetAssetFile, version);

			if (addToWatchList)
				_libWatcher.AddAssetToWatch(asset, targetAssetFile);

		}

        private T LoadFromCache<T>(AssetFile assetFile, bool addToWatchList) where T : Asset, new()
		{
			T asset = null;

            Stream stream = this.OpenReadStream(assetFile.CachedAssetFileFullName);

			if (stream == null) // the library does not contain given file
				return null;

			// input stream should be long enough...
			if (stream.Length < 12)
			{
				stream.Close();
				stream.Dispose();

				return null;
			}

			using (BinaryReader reader = new BinaryReader(stream))
			{
				AssetBinaryReader assetReader = new AssetBinaryReader(reader, this);

				string assetTypeName = reader.ReadString();

				if (typeof(T).Name != assetTypeName)
					return null; // TODO: this is very simple type check ...

				// read asset contents in binary format
				assetReader.VisitAsset<T>(ref asset, null);
			}


			if (asset == null)
				return null;

            asset.SetDependency(this, assetFile.CachedAssetFileFullName);

			WatchAsset(asset, addToWatchList);

			return asset;

		}

		private void SaveToCache<T>(T asset, string file, bool addToWatchList) where T : Asset, new()
		{
            if (asset == null)
                return;

			Stream stream = this.OpenWriteStream(file);
			DateTime version = DateTime.Now;

			// if the version was not set by asset importer we will set it now
			if (asset.Version.Ticks == 0L)
				asset.Version = version;

			using (BinaryWriter writer = new BinaryWriter(stream))
			{
				AssetBinaryWriter assetWriter = new AssetBinaryWriter(writer, this);

				string assetTypeName = typeof(T).Name;

				writer.Write(assetTypeName);

				// save asset contents in binary format
				assetWriter.VisitAsset<T>(ref asset, null);

			}

			SetStreamVersion(file, version);

			if (addToWatchList)
				_libWatcher.AddAssetToWatch(asset, file);

		}

        public T Generate<T>(AssetGenerator generator, bool useCache) where T : Asset, new()
        {
            T result = null;

            if (generator == null)
                return null;

            AssetFile assetFile = GetAssetFileInfo(generator.CachedAssetName);

            if (!string.IsNullOrEmpty(assetFile.AssetName) && useCache)
            {
                if (assetFile.CacheFreshEnough && UseFileCache)
                    return LoadFromCache<T>(assetFile, false);
            }

            result = generator.Generate(this) as T;

            if (string.IsNullOrEmpty(result.Name))
                result.Name = generator.CachedAssetName;

            result.Version = DateTime.Now;

            if (!string.IsNullOrEmpty(assetFile.AssetName) && useCache && UseFileCache)
            {
                SaveToCache<T>(result, assetFile.CachedAssetFileName, false);
            }


            return result;
        }

        /// <summary>
        /// Gets the primary file extension for the specified importer
        /// </summary>
        /// <param name="importer"></param>
        /// <returns></returns>
        public string GetExtension(IAssetImporter importer)
        {
            string[] parts = importer.FileExtension.Split(';');

            if (parts[0].StartsWith("."))
                return parts[0];

            return "." + parts[0];
        }

        public AssetFile GetAssetFileInfo(string assetSource)
        {
            return new AssetFile(this, assetSource);
        }

		/// <summary>
		/// Checks the stream for its version.
		/// </summary>
		/// <param name="assetFileName">The name of the asset being check.</param>
		/// <returns>The date of last write time to the checked stream</returns>
		public virtual DateTime GetStreamVersion(string assetFileName)
		{
			string filePath = GetFileFullPath(assetFileName);

			if (!File.Exists(filePath))
				return DateTime.MinValue;

			return File.GetLastWriteTime(filePath);
		}

		/// <summary>
		/// Sets the version (last write time) for the stream.
		/// </summary>
		/// <param name="assetFileName">The name of the asset file being set.</param>
		public virtual void SetStreamVersion(string assetFileName, DateTime version)
		{
			string filePath = GetFileFullPath(assetFileName);

			if (!File.Exists(filePath))
				return;

			File.SetLastWriteTime(filePath, version);
		}

		/// <summary>
		/// Opens a stream for reading. 
		/// Derived classes can replace this to implement pack files or asset compression.
		/// </summary>
		/// <param name="assetFileName">The name of the asset being read.</param>
		/// <returns>The opened stream.</returns>
		public virtual Stream OpenReadStream(string assetFileName)
		{
			string filePath = GetFileFullPath(assetFileName);

			if (!File.Exists(filePath))
				return null; // nothing to open

			Stream stream;
			try
			{
				Stream fs = new FileStream(Path.Combine(this._rootDirectory.FullName, assetFileName), FileMode.Open, FileAccess.Read, FileShare.Read);

				byte[] dataBuf = new byte[fs.Length];

				fs.Read(dataBuf, 0, dataBuf.Length);

				fs.Close();
				fs.Dispose();

				stream = new MemoryStream(dataBuf);
				//stream = new BufferedStream(fs, StreamBufferSize);
			}
			catch (Exception exception)
			{
				if ((exception is FileNotFoundException) || (exception is DirectoryNotFoundException))
				{
					throw new Exception("Open Stream Not Found: " + assetFileName, exception);
				}
				if (((exception is ArgumentException) || (exception is NotSupportedException)) || ((exception is IOException) || (exception is UnauthorizedAccessException)))
				{
					throw new Exception("Open Stream Error: " +assetFileName, exception);
				}
				throw;
			}
			return stream;
		}

		/// <summary>
		/// Opens a stream for writing. 
		/// Derived classes can replace this to implement pack files or asset compression.
		/// </summary>
		/// <param name="assetFileName">The name of the asset being write.</param>
		/// <returns>The opened stream.</returns>
		public virtual Stream OpenWriteStream(string assetFileName)
		{
			string filePath = GetFileFullPath(assetFileName);

			Stream stream;
			try
			{
				Stream fs = new FileStream(Path.Combine(this._rootDirectory.FullName, assetFileName), FileMode.Create, FileAccess.Write, FileShare.Read);

				stream = new BufferedStream(fs, StreamBufferSize);

			}
			catch (Exception exception)
			{
				if ((exception is FileNotFoundException) || (exception is DirectoryNotFoundException))
				{
					throw new Exception("Open Stream Not Found: " + assetFileName, exception);
				}
				if (((exception is ArgumentException) || (exception is NotSupportedException)) || ((exception is IOException) || (exception is UnauthorizedAccessException)))
				{
					throw new Exception("Open Stream Error: " + assetFileName, exception);
				}
				throw;
			}
			return stream;
		}

		public bool CheckFileExists(string file)
		{
			string filePath = GetFileFullPath(file);

			if (File.Exists(filePath))
				return true;

			return false;

		}

		/// <summary>
		/// Called when a file or directory in the specified System.IO.FileSystemWatcher.Path is changed.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		protected internal virtual void WatcherFileChanged(object sender, FileSystemEventArgs e)
		{
		}

		/// <summary>
		/// Gets the full path for content folder
		/// </summary>
		public string RootFolderPath
		{
			get { return _rootDirectory.FullName; }
		}

		public virtual bool WatchFiles
		{
			get
			{
				return _libWatcher.WatchFiles;
			}
			set
			{
				_libWatcher.WatchFiles = value;
			}
		}

		public void ResetImporter()
		{
			
		}

		public void ResetExporter()
		{
		}


		public string FileExtension
		{
			get { return ".asset"; }
		}

		public bool HasParentImporter
		{
			get { return false; }
		}

		public AssetLibrary Library
		{
			get
			{
				return this;;
			}
			set
			{
			}
		}


	}

}
