﻿using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace XnaDevRu.Assets
{
	public class AssetRender : IDisposable
	{
		internal class DipAdapter
		{
			GraphicsDevice _device;
			bool _useStreams;
			bool _safeMode;

			internal DipAdapter(GraphicsDevice device, bool useStreams, bool safeMode)
			{
				_device = device;
				_safeMode = safeMode;
				_useStreams = useStreams;
				
				if (safeMode)
					_useStreams = false;
			}

			internal bool BindVertexData(ModelAsset model, StreamDeclaration decl, bool update)
			{
				if (model == null)
					return false;

				if (model.Meshes.Count == 0)
					return false;

				if (_safeMode) // we can`t use vertex streams in safe mode
					decl.BindTo(_device, false);
				else
					decl.BindTo(_device, _useStreams);

				if (_safeMode)
					return true;

				if (_useStreams)
				{
					_device.Vertices[0].SetSource(model.GetVertexPosBuffer(_device, update), 0, 12);
					_device.Vertices[1].SetSource(model.GetVertexNormBuffer(_device, update), 0, 12);
					_device.Vertices[2].SetSource(model.GetVertexTexBuffer(_device, update), 0, 8);
				}
				else
					_device.Vertices[0].SetSource(model.GetVertexBuffer(_device, update), 0, model.VertexStride);

				_device.Indices = model.GetIndexBuffer(_device, update);

				return true;
			}

			internal void Dip(ModelAsset model, MeshPartAsset meshPart)
			{
				int numVertices = model.VertexCount;
				int startIndex = meshPart.StartIndex;
				int primCount = meshPart.PrimitiveCount;
				PrimitiveType pType = meshPart.PrimitiveTopology;

				if (!_safeMode)
				{
					_device.DrawIndexedPrimitives(pType, 0, 0, numVertices, startIndex, primCount);
				}
				else
				{
					_device.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(pType, model.VertexData, 0, numVertices, model.IndexData, 0, primCount);
				}
			}

		}

		GraphicsDevice _device;
        SpriteBatch _batch;
        FontAsset font;
		StreamDeclaration _PNTDecl;
		StreamDeclaration _PDecl;

		RenderTarget2D rtFront;
		RenderTarget2D rtBack;
		DipAdapter dipAdapter;


		public AssetRender(GraphicsDevice device, bool useStreams, bool safeMode)
		{
			this._device = device;
			_batch = new SpriteBatch(device);

			_PNTDecl = new StreamDeclaration(VertexPositionNormalTexture.VertexElements);
			_PDecl = new StreamDeclaration(VertexPositionNormalTexture.VertexElements[0]);

			rtFront = new RenderTarget2D(device, 512, 512, 0, device.DisplayMode.Format);
			rtBack = new RenderTarget2D(device, 512, 512, 0, device.DisplayMode.Format);

			dipAdapter = new DipAdapter(device, useStreams, safeMode);

		}

		public void RenderModel(ModelAsset model, LambertShader shader, bool update)
		{
			if (model == null || shader == null)
				return;

			if (!dipAdapter.BindVertexData(model, _PNTDecl, update))
				return;

			for (int meshIdx = 0; meshIdx < model.Meshes.Array.Length; meshIdx++)
			{
				MeshAsset mesh = model.Meshes.Array[meshIdx];

				shader.World = mesh.WorldMatrix;

				for (int i = 0; i < mesh.MeshParts.Array.Length; i++)
				{
					MeshPartAsset meshPart = mesh.MeshParts.Array[i];

					MaterialAsset material = meshPart.Material;

					Texture2D texture = material.ResolveTexture(_device) as Texture2D;

					shader.DiffuseLight = material.DiffuseColor.ToVector3();

					if (texture == null)
					{
						shader.TextureEnabled = false;
					}
					else
					{
						shader.TextureEnabled = true;
						shader.Texture = texture;
					}

					//effect.CommitChanges();

					shader.Begin(true);

					dipAdapter.Dip(model, meshPart);

                    shader.End(true);

				}
			}


		}

		public void RenderFogModel(ModelAsset model, FogShader shader, bool useStreams, bool update)
		{
			if (model == null || shader == null)
				return;

			if (model.Meshes.Count == 0)
				return;

			_PDecl.BindTo(_device, useStreams);

			if (!useStreams)
			{
				_device.Vertices[0].SetSource(model.GetVertexPosBuffer(_device, update), 0, 12);
			}
			else
			{
				_device.Vertices[0].SetSource(model.GetVertexPosBuffer(_device, update), 0, 12);
			}
			_device.Indices = model.GetIndexBuffer(_device, update);

			for (int meshIdx = 0; meshIdx < model.Meshes.Array.Length; meshIdx++)
			{
				MeshAsset mesh = model.Meshes.Array[meshIdx];

				for (int i = 0; i < mesh.MeshParts.Array.Length; i++)
				{
					MeshPartAsset meshPart = mesh.MeshParts.Array[i];

					MaterialAsset material = meshPart.Material;

					Texture2D texture = material.ResolveTexture(_device) as Texture2D;

					int numVertices = model.VertexCount;
					int startIndex = meshPart.StartIndex;
					int primCount = meshPart.PrimitiveCount;
					PrimitiveType pType = meshPart.PrimitiveTopology;

					_device.SetRenderTarget(0, rtBack);
					_device.Clear(Color.Black);

					shader.BeginBack();

					shader.BeginPass(0);

					_device.DrawIndexedPrimitives(pType, 0, 0, numVertices, startIndex, primCount);

					shader.EndPass(0);
					shader.End();

					_device.ResolveRenderTarget(0);
					Texture2D bTex = rtBack.GetTexture() as Texture2D;
					//bTex.Save("back.dds", ImageFileFormat.Dds);

					_device.SetRenderTarget(0, rtFront);
					_device.Clear(Color.Black);

					shader.BeginFront();

					shader.BeginPass(0);

					_device.DrawIndexedPrimitives(pType, 0, 0, numVertices, startIndex, primCount);

					shader.EndPass(0);
					shader.End();

					_device.ResolveRenderTarget(0);
					Texture2D fTex = rtFront.GetTexture() as Texture2D;
					//fTex.Save("front.dds", ImageFileFormat.Dds);

					_device.SetRenderTarget(0, null);

					shader.FrontTexture = fTex;
					shader.BackTexture = bTex;

					shader.BeginTexturing();

					shader.BeginPass(0);

					_device.DrawIndexedPrimitives(pType, 0, 0, numVertices, startIndex, primCount);

					shader.EndPass(0);
					shader.End();

				}
			}


		}

        public void BeginText(FontAsset font, SpriteBlendMode blend)
        {
            if (this.font != null)
                return; // error, call EndText() first..

            this.font = font;

            if (this.font == null)
                return;

            // the fastest possible - Immediate, Sprite Sort Mode
            this._batch.Begin(blend, SpriteSortMode.Immediate, SaveStateMode.None);
        }

        public void EndText()
        {
            if (this.font == null)
                return;

            this.font = null;
            this._batch.End();

        }


        /// <summary>
        /// Draws text to the screen.
        /// </summary>
        public void DrawText(string text, Vector2 position, float scale, Color color)
        {
            if (this.font == null)
                return; // call BeginText(..) first..

            if (string.IsNullOrEmpty(text))
                return; // nothing to draw..

            Texture2D tex2D = font.Resolve(_device);

            if(tex2D == null)
                return; // font was not created correctly..

            Rectangle glyphRect = new Rectangle();

            for(int i=0;i<text.Length; i++)
            {
                char ch = text[i];

                // Look up what font glyph corresponds to this Unicode character.

                int index = font.CharTable[(int)ch];

                if (index == 0)
                    continue;

                glyphRect.X = font.GlyphPosXAarray[index];
                glyphRect.Y = font.GlyphPosYAarray[index];

                glyphRect.Width = font.GlyphWidthAarray[index];
                glyphRect.Height = font.GlyphHeightAarray[index];

                // Draw the character.

                _batch.Draw(tex2D, position, glyphRect, color, 0,
                                 Vector2.Zero, scale, SpriteEffects.None, 0);

                // Move the position further to the right.
                position.X += glyphRect.Width * scale;
            }
        }

        #region IDisposable Members

        public void Dispose()
        {
            _batch.Dispose();
        }

        #endregion
    }
}
