﻿using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XnaDevRu.Assets
{
	public class ModelAsset : Asset<ModelAsset>
	{
		/// <summary>
		/// Array of NodeAsset`s which describe how each mesh in the ModelAsset.Meshes array for this model relates to its parent mesh
		/// </summary>
		public DataList<NodeAsset> Nodes;

		/// <summary>
		/// Array of MeshAsset`s which compose the model. 
		/// Each MeshAsset in a model may be moved independently and may be composed of multiple materials 
		/// identified as MeshPartAsset objects.
		/// </summary>
		public DataList<MeshAsset> Meshes;
		public MaterialPackAsset Materials;

		public Vector3[] PositionsArray;
		public Vector3[] NormalsArray;
		public Vector2[] TexCoordsArray;
		public Vector3[] TangentsArray;
		public Vector3[] BitangentsArray;
		public Color[] ColorsArray;

        BoundingSphere? bSphere;

		private VertexBuffer _vBuffer;
		private VertexBuffer _vPosBuffer;
		private VertexBuffer _vNormBuffer;
		private VertexBuffer _vTexBuffer;
		private IndexBuffer _iBuffer;
        private VertexPositionNormalTexture[] cachedVBufData;
		private int[] cachedIBufData;

		public ModelAsset()
		{
		}

		public override void Accept(Accessor visitor)
		{
			// visit Name and Version fields ..
			base.Accept(visitor);

			visitor.VisitSubAssetArray<NodeAsset>(ref Nodes, "Nodes", this);
			visitor.VisitSubAssetArray<MeshAsset>(ref Meshes, "Meshes", this);
			visitor.VisitAsset<MaterialPackAsset>(ref Materials, "MaterialPack");

			visitor.VisitArray(ref cachedIBufData, "IndexBuffer");

			visitor.VisitArray(ref PositionsArray, "Positions");
			visitor.VisitArray(ref NormalsArray, "Normals");
			visitor.VisitArray(ref TexCoordsArray, "TexCoords");
			visitor.VisitArray(ref TangentsArray, "Tangents");
			visitor.VisitArray(ref BitangentsArray, "Bitangents");
			visitor.VisitArray(ref ColorsArray, "Colors");
		}

		#region runtime asset construction methods

		/// <summary>
		/// Sets the geometry and indices data
		/// </summary>
		/// <param name="geomData">vertices array</param>
		/// <param name="indexData">indices array (default is CCW order)</param>
		/// <param name="swapIndexOrder">true to change index order (convert CW order to CCW)</param>
		public void SetGeometry(VertexPositionNormalTexture[] geomData, int[] indexData, bool swapIndexOrder)
		{
			cachedIBufData = indexData;

			if (swapIndexOrder)
				SwapIndexOrder();

			if (geomData == null)
				return;

			TangentsArray = null;
			BitangentsArray = null;

			PositionsArray = new Vector3[geomData.Length];
			NormalsArray = new Vector3[geomData.Length];
			TexCoordsArray = new Vector2[geomData.Length];

			for (int i = 0; i < geomData.Length; i++)
			{
				PositionsArray[i] = geomData[i].Position;
				NormalsArray[i] = geomData[i].Normal;
				TexCoordsArray[i] = geomData[i].TextureCoordinate;
			}
		}

		public NodeAsset DefineNode(Matrix matrix)
		{
			NodeAsset asset = new NodeAsset(this, matrix);
			asset.Index = Nodes.Count;

			Nodes.Add(asset);

			return asset;
		}

		public MeshAsset DefineMesh(NodeAsset node)
		{
			MeshAsset asset = new MeshAsset(this);
			asset.NodeIndex = node.Index;

			Meshes.Add(asset);

			return asset;
		}

		public MaterialAsset DefineMaterial2D(Color diffuse, string textureName)
		{
			// material with external texture

			if (Materials == null)
				Materials = new MaterialPackAsset();

			MaterialAsset asset = Materials.DefineMaterial(null);
			asset.DiffuseColor = diffuse;
			asset.Texture = new TextureAsset(textureName, TextureAsset.Type.Texture2D);

			return asset;
		}

		public MaterialAsset DefineMaterial()
		{
			if (Materials == null)
				Materials = new MaterialPackAsset();

			MaterialAsset asset = Materials.DefineMaterial(null);

			return asset;
		}

		public MaterialAsset DefineMaterial2D(Color diffuse, Texture2D texture)
		{
			// material with embedded texture

			if (Materials == null)
				Materials = new MaterialPackAsset();

			MaterialAsset asset = Materials.DefineMaterial(null);
			asset.DiffuseColor = diffuse;

			// TODO implement embedded texture definition

			return asset;
		} 

		#endregion

		internal void OptimizeVertexData()
		{
			if (PositionsArray == null)
				return; // nothing to optimize

			if (cachedIBufData == null)
			{
				cachedIBufData = new int[PositionsArray.Length];

				for (int i = 0; i < cachedIBufData.Length; i++)
					cachedIBufData[i] = i;
			}

			int[] tt = new int[PositionsArray.Length];	// translation table

			List<int> uniqueVertexList = new List<int>(PositionsArray.Length / 2);

			int max = 0;

			for (int i = 0; i < PositionsArray.Length; i++)
			{
				tt[i] = -1;

				int index = cachedIBufData[i];

				int newIdx = FindVertex(i, uniqueVertexList);

				if (newIdx == -1) // we have no such vertex yet
				{
					tt[index] = uniqueVertexList.Count;

					uniqueVertexList.Add(index); // add unique vertex`s index to the list

					max = i;
					continue;
				}

				tt[index] = newIdx;

			}

			// loop through unique vetices

			Vector3[] newPosArr = new Vector3[uniqueVertexList.Count];

			Vector3[] newNormArr = null;
			Vector2[] newTexCArr = null;
			Vector3[] newTangArr = null;
			Vector3[] newBitanArr = null;
			Color[] newColorArr = null;

			if (NormalsArray != null)
				newNormArr = new Vector3[uniqueVertexList.Count];

			if (TexCoordsArray != null)
				newTexCArr = new Vector2[uniqueVertexList.Count];

			if (TangentsArray != null)
				newTangArr = new Vector3[uniqueVertexList.Count];

			if (BitangentsArray != null)
				newBitanArr = new Vector3[uniqueVertexList.Count];

			if (ColorsArray != null)
				newColorArr = new Color[uniqueVertexList.Count];

			for (int i = 0; i < newPosArr.Length; i++)
			{
				newPosArr[i] = PositionsArray[uniqueVertexList[i]];

				if (NormalsArray != null)
					newNormArr[i] = NormalsArray[uniqueVertexList[i]];

				if (TexCoordsArray != null)
					newTexCArr[i] = TexCoordsArray[uniqueVertexList[i]];

				if (TangentsArray != null)
					newTangArr[i] = TangentsArray[uniqueVertexList[i]];
				
				if (BitangentsArray != null)
					newBitanArr[i] = BitangentsArray[uniqueVertexList[i]];

				if (ColorsArray != null)
					newColorArr[i] = ColorsArray[uniqueVertexList[i]];

			}

			PositionsArray = newPosArr;
			NormalsArray = newNormArr;
			TexCoordsArray = newTexCArr;
			TangentsArray = newTangArr;
			BitangentsArray = newBitanArr;
			ColorsArray = newColorArr;

			cachedIBufData = tt;

		}

		private int FindVertex(int index, List<int> uniqueVertexList)
		{
			for (int i = 0; i < uniqueVertexList.Count; i++)
			{
				int idx = uniqueVertexList[i];

				if (PositionsArray[idx] != PositionsArray[index])
					continue;

				if (NormalsArray != null)
					if (NormalsArray[idx] != NormalsArray[index])
						continue;

				if (TexCoordsArray != null)
					if (TexCoordsArray[idx] != TexCoordsArray[index])
						continue;

				if (TangentsArray != null)
					if (TangentsArray[idx] != TangentsArray[index])
						continue;

				if (BitangentsArray != null)
					if (BitangentsArray[idx] != BitangentsArray[index])
						continue;

				if (ColorsArray != null)
					if (ColorsArray[idx] != ColorsArray[index])
						continue;

				return i;

			}

			return -1;
		}

		public VertexBuffer GetVertexBuffer(GraphicsDevice device, bool update)
		{
            VertexPositionNormalTexture[] buf = VertexData;
            
            if (_vBuffer == null)
			{
				_vBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * buf.Length, ResourceUsage.None);

				_vBuffer.SetData<VertexPositionNormalTexture>(buf);
			}
            else
                if(update)
                    _vBuffer.SetData<VertexPositionNormalTexture>(buf);

			return _vBuffer;
		}

		public VertexBuffer GetVertexPosBuffer(GraphicsDevice device, bool update)
		{
			if (_vPosBuffer == null)
			{
				_vPosBuffer = new VertexBuffer(device, 12 * this.PositionsArray.Length, ResourceUsage.None);

				_vPosBuffer.SetData<Vector3>(this.PositionsArray);
			}
            else
                if(update)
                    _vPosBuffer.SetData<Vector3>(this.PositionsArray);

			return _vPosBuffer;
		}

		public VertexBuffer GetVertexNormBuffer(GraphicsDevice device, bool update)
		{
			if (_vNormBuffer == null)
			{
				_vNormBuffer = new VertexBuffer(device, 12 * this.NormalsArray.Length, ResourceUsage.None);

				_vNormBuffer.SetData<Vector3>(this.NormalsArray);
			}
            else
                if (update)
                    _vNormBuffer.SetData<Vector3>(this.NormalsArray);

			return _vNormBuffer;
		}

		public VertexBuffer GetVertexTexBuffer(GraphicsDevice device, bool update)
		{
			if (_vTexBuffer == null)
			{
				_vTexBuffer = new VertexBuffer(device, 8 * this.TexCoordsArray.Length, ResourceUsage.None);

				_vTexBuffer.SetData<Vector2>(this.TexCoordsArray);
			}
            else
                if(update)
                    _vTexBuffer.SetData<Vector2>(this.TexCoordsArray);

			return _vTexBuffer;
		}

		public IndexBuffer GetIndexBuffer(GraphicsDevice device, bool update)
		{

			if (_iBuffer == null)
			{
				_iBuffer = new IndexBuffer(device, 4 * IndexData.Length, ResourceUsage.None, IndexElementSize.ThirtyTwoBits);

				_iBuffer.SetData<int>(IndexData);
			}
            else
                if(update)
                    _iBuffer.SetData<int>(IndexData);


			return _iBuffer;
		}

		public VertexPositionNormalTexture[] VertexData
		{
			get
			{
                if (cachedVBufData == null)
                {
                    if (PositionsArray == null)
                        return null;

                    VertexPositionNormalTexture[] result = new VertexPositionNormalTexture[PositionsArray.Length];

                    for (int i = 0; i < result.Length; i++)
                    {
                        result[i].Position = PositionsArray[i];

                        if (NormalsArray != null)
                            result[i].Normal = NormalsArray[i];
                        else
                            result[i].Normal = Vector3.Up;

                        if (TexCoordsArray != null)
                            result[i].TextureCoordinate = TexCoordsArray[i];
                    }

                    cachedVBufData = result;
                }

                return cachedVBufData;
			}
		}

		public int[] IndexData
		{
			get
			{

				if (cachedIBufData == null)
				{
					if (PositionsArray == null)
						return null;

					cachedIBufData = new int[PositionsArray.Length];

					for (int i = 0; i < cachedIBufData.Length; i++)
						cachedIBufData[i] = i;

				}

				return cachedIBufData;
			}
		}

		private void SwapIndexOrder()
		{
			for (int i = 2; i < cachedIBufData.Length; i += 3)
			{
				int tmp = cachedIBufData[i - 2];
				IndexData[i - 2] = cachedIBufData[i];
				cachedIBufData[i] = tmp;
			}
		}

		public bool UseVertexColors
		{
			get
			{
				return ColorsArray != null;
			}
		}

		public int VertexStride
		{
			get
			{
				return VertexPositionNormalTexture.SizeInBytes;
			}
		}

		public int VertexCount
		{
			get
			{
				if (PositionsArray == null)
					return 0;

				return PositionsArray.Length;
			}
		}

        public BoundingSphere BoundingSphere
        {
            get
            {
                if (PositionsArray == null)
                    new BoundingSphere();

                if (bSphere.HasValue)
                    return bSphere.Value;

                bSphere = Microsoft.Xna.Framework.BoundingSphere.CreateFromPoints(PositionsArray);

                return bSphere.Value;
            }
        }


	}
}
