﻿using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework.Graphics;
using System.IO;

namespace XnaDevRu.Assets
{
	public class TextureAsset : Asset<TextureAsset>
	{
		// the data source for texture content
		TextureSource _source;

		/// <summary>
		/// Defines texture types.
		/// </summary>
		public enum Type : byte
		{
			/// <summary>An unknown texture resource.</summary>
			Unknown = 0, // null
			/// <summary>A 2-dimensional texture resource.</summary>
			Texture2D = 1,
			/// <summary>A 3-dimensional texture resource.</summary>
			Texture3D = 2,
			/// <summary>A cube texture resource.</summary>
			TextureCube = 3,

		}

		internal enum SourceType : byte
		{
			/// <summary>An unknown texture source.</summary>
			Unknown = 0, // null
			/// <summary>An external texture file.</summary>
			External = 1,
			/// <summary>An embedded texture resource.</summary>
			Embedded = 2,

		}

		public Type TextureType;

		public TextureAsset() { }

		public TextureAsset(string file, TextureAsset.Type tType)
		{
			TextureType = tType;

			_source = new ExternalTextureSource(file);

		}

        public TextureAsset(Stream stream, TextureAsset.Type tType)
        {
            TextureType = tType;

            _source = new EmbeddedTextureSource(stream);

        }


		public override void Accept(Accessor visitor)
		{
            base.Accept(visitor);

			byte type = (byte)TextureType;

			visitor.Visit(ref type, "TextureType");

			TextureType = (Type)type;
			
			// read/write the code to identify the type of the source
			// could be External or Embedded at this moment

			ExternalTextureSource extSrc = _source as ExternalTextureSource;
			EmbeddedTextureSource embSrc = _source as EmbeddedTextureSource;

			byte srcTypeCode = (int)SourceType.Unknown; // unknown or null

			if (extSrc != null)
				srcTypeCode = (int)SourceType.External;

			if (embSrc != null)
				srcTypeCode = (int)SourceType.Embedded;

			visitor.Visit(ref srcTypeCode, "TextureSourceType");

			SourceType srcType = (SourceType)srcTypeCode;

			if (srcType == SourceType.Unknown) // unknown or null
			{
				return;
			}

			if (srcType == SourceType.External)
			{
				if (extSrc == null)
				{
					extSrc = new ExternalTextureSource();
					_source = extSrc;
				}

				extSrc.Accept(visitor);

			}

			if (srcType == SourceType.Embedded)
			{
				if (embSrc == null)
				{
					embSrc = new EmbeddedTextureSource();
					_source = embSrc;
				}

				embSrc.Accept(visitor);

			}



		}

		public Texture Resolve(GraphicsDevice device)
		{
			if (_source == null)
				return null;

			if (TextureType == TextureAsset.Type.Unknown)
				return null;

			return _source.Resolve(device, TextureType);
		}

        public void SetSource(TextureAsset texture)
        {
            if (texture == null)
                return;

            if (texture._source == null)
                return;

            if (this._source != null)
            {
                this._source.Dispose();
            }

            this._source = texture._source;

        }

	}

	abstract class TextureSource : IDisposable
	{
		protected Texture _texture;

		protected internal abstract Texture Resolve(GraphicsDevice device, TextureAsset.Type type);

		protected internal abstract void Accept(Accessor visitor);


        #region IDisposable Members

        public void Dispose()
        {
            if (_texture == null)
                return;

            if (!_texture.IsDisposed)
                _texture.Dispose();

            _texture = null;

        }

        #endregion
    }

	class ExternalTextureSource : TextureSource
	{
		public string FileName;

		public ExternalTextureSource() { }

		public ExternalTextureSource(string file) 
		{
			FileName = file;
		}

		protected internal override Texture Resolve(GraphicsDevice device, TextureAsset.Type type)
		{
			if (_texture != null)
				return _texture;

            using (FileStream stream = new FileStream(FileName, FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                byte[] buf = new byte[stream.Length];
                stream.Read(buf, 0, buf.Length);

                using (MemoryStream ms = new MemoryStream(buf))
                {
                    _texture = Texture.FromFile(device, ms);
                }
            }

			return _texture;
		}

		protected internal override void Accept(Accessor visitor)
		{
			visitor.Visit(ref FileName, "Texture");
		}

    }

	// TODO test it ...
	class EmbeddedTextureSource : TextureSource
	{
		internal enum EmbedType : byte
		{
			Unknown = 0,
			Array = 1,
			Stream = 2,
		}

		private EmbedType Embedding;
		private int Width; // for 2D textures
		private int Height; // for 2D textures
		private int Depth; // for 3D textures
		private int EdgeSize; // for cube textures

		private SurfaceFormat Format;
		private byte[] Data;
		//private 


		public EmbeddedTextureSource() { }

        public EmbeddedTextureSource(Stream sourceStream) 
        {
            TextureInformation tInfo = Texture.GetTextureInformation(sourceStream);

            this.Embedding = EmbedType.Stream;
            this.Width = tInfo.Width;
            this.Height = tInfo.Height;
            this.Depth = tInfo.Depth;
            this.EdgeSize = -1;
            this.Format = tInfo.Format;

            this.Data = new byte[sourceStream.Length];

            sourceStream.Position = 0;

            sourceStream.Read(this.Data, 0, Data.Length);
 
        }

		protected internal override Texture Resolve(GraphicsDevice device, TextureAsset.Type type)
		{
			if (_texture != null)
				return _texture;

			if (Embedding == EmbedType.Unknown)
				return null;

			if (Embedding == EmbedType.Array)
			{
				if (type == TextureAsset.Type.Texture2D && Data != null)
				{
					Texture2D tex = new Texture2D(device, Width, Height, 0, ResourceUsage.None | ResourceUsage.AutoGenerateMipMap, Format);

					tex.SetData<byte>(Data);

					_texture = tex;
				}

				if (type == TextureAsset.Type.Texture3D && Data != null)
				{
					Texture3D tex = new Texture3D(device, Width, Height, Depth, 0, ResourceUsage.None | ResourceUsage.AutoGenerateMipMap, Format);

					tex.SetData<byte>(Data);

					_texture = tex;
				}

				if (type == TextureAsset.Type.TextureCube && Data != null)
				{
					TextureCube tex = new TextureCube(device, EdgeSize, 0, ResourceUsage.None | ResourceUsage.AutoGenerateMipMap, Format);

					// TODO implement SetData for cube textures
					//tex.SetData<byte>(Data);

					_texture = tex;
				}


				return _texture;
			}

			if (Embedding == EmbedType.Stream && Data != null)
			{
                using (MemoryStream stream = new MemoryStream(Data))
                {

                    _texture = Texture.FromFile(device, stream);
                }

			}


			return _texture;
		}

		protected internal override void Accept(Accessor visitor)
		{
			byte embedTypeCode = (byte)Embedding;

			visitor.Visit(ref embedTypeCode, "Embedding");

			Embedding = (EmbedType)embedTypeCode;

			visitor.Visit(ref Width, "Width");
			visitor.Visit(ref Height, "Height");
			visitor.Visit(ref Depth, "Depth");
			visitor.Visit(ref EdgeSize, "EdgeSize");

			int formatTypeCode = (int)Format;

			visitor.Visit(ref formatTypeCode, "Format");

			Format = (SurfaceFormat)formatTypeCode;

			visitor.VisitArray(ref Data, "Data");
		}
	}

}
