﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;

namespace XnaDevRu.Assets
{
	sealed internal class LibraryWatcher
	{
		#region fields
		private AssetLibrary _library;

		private FileSystemWatcher _watcher;
		private FileSystemEventHandler _watcherHandler;
		private List<Asset> _watchedAssetList;
		private Dictionary<string, Asset> _assetFileToAssetInstanceMap;
		private List<string> _changedFileList;
		private List<string> _watchedFileList;
		private Thread _worker;
		private bool _isBusy;
		private string _currentWorkingFileName;
		#endregion

		internal LibraryWatcher(AssetLibrary lib)
		{
			_library = lib;

			_watcherHandler = new FileSystemEventHandler(this.WatcherFileChanged);
			_watchedFileList = new List<string>();

		}

		internal void AddAssetDependencyToWatch(Asset asset)
		{
			if (asset.DependencyList == null)
				return;

			for (int i = 0; i < asset.DependencyList.Length; i++)
				this.AddAssetToWatch(asset, asset.DependencyList[i]);
		}

		internal void AddAssetToWatch(Asset asset, string file)
		{
			if (asset == null || string.IsNullOrEmpty(file))
				return;

			if (_watchedAssetList == null)
				return;

			string fileFullPath = _library.GetFileFullPath(file);

			if (string.IsNullOrEmpty(fileFullPath))
				return;

            Uri uri = new Uri(fileFullPath);

			if (!_watchedAssetList.Contains(asset))
				_watchedAssetList.Add(asset);

			if (!_watchedFileList.Contains(fileFullPath))
				_watchedFileList.Add(fileFullPath);

			if (!_assetFileToAssetInstanceMap.ContainsKey(uri.AbsolutePath))
				_assetFileToAssetInstanceMap.Add(uri.AbsolutePath, asset);
		}

		// called from worker thread
		private void WorkerMethod()
		{
			while (true)
			{
				if (_isBusy)
				{
					Thread.Sleep(50);
					continue;
				}

				if (_changedFileList == null)
				{
					Thread.Sleep(300);
					continue;
				}

				if (_changedFileList.Count == 0)
				{
					Thread.Sleep(100);
					continue;
				}

				string[] changedFileFullNames = null;

				lock (_changedFileList)
				{
					changedFileFullNames = _changedFileList.ToArray();
					_changedFileList.Clear();
				}

				for (int i = 0; i < changedFileFullNames.Length; i++)
				{
					string changedFile = changedFileFullNames[i];
					Asset oldAsset = null;

					if (_assetFileToAssetInstanceMap.TryGetValue(changedFile, out oldAsset))
					{
						if (oldAsset != null)
						{
							Asset newAsset = oldAsset.ReloadAsset(this._library, changedFile);

							if (newAsset != null)
								oldAsset.Recreate(this._library, newAsset);

							Thread.Sleep(50);

						}
					}
				}


			}
		}

		private void CreateFileWatcher()
		{
			if (_watcher != null)
				return; // we are watching already

			_watcher = new FileSystemWatcher(_library.RootFolderPath);
			_watcher.NotifyFilter = NotifyFilters.LastWrite;
			_watcher.EnableRaisingEvents = true;
			_watcher.IncludeSubdirectories = true;
			_watcher.Changed += _watcherHandler;

			if (_watchedAssetList != null)
				_watchedAssetList.Clear();
			else
				_watchedAssetList = new List<Asset>();

			if (_assetFileToAssetInstanceMap != null)
				_assetFileToAssetInstanceMap.Clear();
			else
				_assetFileToAssetInstanceMap = new Dictionary<string, Asset>();

			if (_changedFileList != null)
				_changedFileList.Clear();
			else
				_changedFileList = new List<string>();

			if (_worker == null)
			{
				_worker = new Thread(new ThreadStart(WorkerMethod));
				_worker.IsBackground = true;
				_worker.Priority = ThreadPriority.Lowest;
				_worker.Start();
			}


		}

		private void DestroyFileWatcher()
		{
			if (_watcher == null)
				return; // we are not watching already

			_watcher.EnableRaisingEvents = false;
			_watcher.Changed -= _watcherHandler;
			_watcher.Dispose();
			_watcher = null;

			if (_watchedAssetList != null)
				_watchedAssetList.Clear();

			if (_assetFileToAssetInstanceMap != null)
				_assetFileToAssetInstanceMap.Clear();

			if (_changedFileList != null)
				_changedFileList.Clear();


			_watchedAssetList = null;
			_assetFileToAssetInstanceMap = null;
			_changedFileList = null;

			if (_worker != null)
			{
				_worker.Abort();
				_worker = null;
			}

		}

		/// <summary>
		/// Called when a file or directory in the specified System.IO.FileSystemWatcher.Path is changed.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		/// <remarks>This method is called form the thread of FileSystemWatcher object</remarks>
		internal void WatcherFileChanged(object sender, FileSystemEventArgs e)
		{
			if (_changedFileList == null)
				return;

			if (e.ChangeType == WatcherChangeTypes.Changed)
			{
				if (!string.IsNullOrEmpty(_currentWorkingFileName))
				{
					if (Path.GetFileName(_currentWorkingFileName) == e.Name)
						return;
				}

				_isBusy = true;

                Uri uri = new Uri(e.FullPath);

				lock (_changedFileList)
				{
					if (!_changedFileList.Contains(uri.AbsolutePath))
                        _changedFileList.Add(uri.AbsolutePath);
				}
			}

			_library.WatcherFileChanged(sender, e);

			_isBusy = false;
		}

		internal List<Asset> WatchedAssetList
		{
			get { return _watchedAssetList; }
		}

		internal bool WatchFiles
		{
			get
			{
				return _watcher != null;
			}
			set
			{
				if (value == true)
				{
					CreateFileWatcher();
				}
				else
					DestroyFileWatcher();
			}
		}

        internal string CurrentWorkingFileName
        {
            set 
            {
                if (string.IsNullOrEmpty(value))
                    _currentWorkingFileName = null;
                else
                    _currentWorkingFileName = new Uri(value).AbsolutePath;
            }
        }


	}
}
